#include <fxlib.h>
#include "thinice.h"
#include "tileset.h"
#include "functions.h"
#include "MonochromeLib.h"

#define FORCE_STEP 0.001
#define FORCE_STEP2 -0.003
#define FORCE_MAX 0.02
#define SPEED_MAX 1
#define ACCEL_MAX 0.01
#define FRICTION 0.02

#define TIME_SKATING 50

#define NB_TRACE 40
#define DIFF_MIN_C 10
#define CPT_MIN 15
#define TIME_HOLE 400
#define TIME_HOLE_MIN 100

#define ENEMY_SIZE_C 200
#define NB_TEST 30

#define TIME_ENEMY 200


void moveSkater(Skater *skt) {
	int time = RTC_getTicks();
	float speed, accel;
	
	speed = skt->speedX*(time-skt->time);
	if(speed > 0 && speed > SPEED_MAX) speed = SPEED_MAX;
	if(speed < 0 && speed < -SPEED_MAX) speed = -SPEED_MAX;
	skt->skaterX += speed;
	if(skt->skaterX < 0) {
		skt->skaterX = 0;
		skt->speedX = -skt->speedX;
	}
	if(skt->skaterX > 117) {
		skt->skaterX = 117;
		skt->speedX = -skt->speedX;
	}
	
	speed = skt->speedY*(time-skt->time);
	if(speed > 0 && speed > SPEED_MAX) speed = SPEED_MAX;
	if(speed < 0 && speed < -SPEED_MAX) speed = -SPEED_MAX;
	skt->skaterY += speed;
	if(skt->skaterY < 0) {
		skt->skaterY = 0;
		skt->speedY = -skt->speedY;
	}
	if(skt->skaterY > 49) {
		skt->skaterY = 49;
		skt->speedY = -skt->speedY;
	}
	
	accel = (skt->forceX-FRICTION*skt->speedX)*(time-skt->time);
	if(accel > 0 && accel > ACCEL_MAX) accel = ACCEL_MAX;
	if(accel < 0 && accel < -ACCEL_MAX) accel = -ACCEL_MAX;
	skt->speedX += accel;
	
	accel = (skt->forceY-FRICTION*skt->speedY)*(time-skt->time);
	if(accel > 0 && accel > ACCEL_MAX) accel = ACCEL_MAX;
	if(accel < 0 && accel < -ACCEL_MAX) accel = -ACCEL_MAX;
	skt->speedY += accel;
	
	skt->time = time;
}

int engine(void) {
	Skater skt;
	Tileset *sktFrontBack = tileset_load(P_FRONT_BACK);
	Tileset *sktSide = tileset_load(P_SIDE);
	Tileset *enemy = tileset_load(ENEMY);
	
	int timeSkating = RTC_getTicks();
	int posSkating = 0;
	int posSkater = 0;
	
	//ice
	int traceX[NB_TRACE], traceY[NB_TRACE];
	int i = 0, iLoop = 0;
	int hole = 0, timeHole;
	int abscissa[NB_TRACE] = {0}, ordinate[NB_TRACE] = {0};
	int i1, i2;
	int j, j1, j2;
	int cpt, cpt2, k, k1;
	int defend1, defend2;
	
	int nbEnemy = 0;
	int enemyX, enemyY;
	int timeEnemy = RTC_getTicks();
	
	skt.skaterX = 58; skt.skaterY = 25;
	skt.speedX = 0; skt.speedY = 0;
	skt.forceX = 0; skt.forceY = 0;
	skt.time = RTC_getTicks();
	
	random(RTC_getTicks(), 1);
	
	do {
		enemyX = 15+random(0, 86);
		enemyY = 16+random(0, 16);
	} while((enemyX-skt.skaterX)*(enemyX-skt.skaterX)+(enemyY-skt.skaterY)*(enemyY-skt.skaterY) < ENEMY_SIZE_C);
	
	while(!IsKeyDown(KEY_CTRL_EXIT)) {
		if(IsKeyDown(KEY_CTRL_RIGHT)) {
			skt.forceX += FORCE_STEP;
			if(skt.forceX > FORCE_MAX) skt.forceX = FORCE_MAX;
		}
		else if(IsKeyDown(KEY_CTRL_LEFT)) {
			skt.forceX -= FORCE_STEP;
			if(skt.forceX < -FORCE_MAX) skt.forceX = -FORCE_MAX;
		}
		else if(skt.forceX > 0) {
			skt.forceX += FORCE_STEP2;
			if(skt.forceX < 0) skt.forceX = 0;
		}
		else if(skt.forceX < 0) {
			skt.forceX -= FORCE_STEP2;
			if(skt.forceX > 0) skt.forceX = 0;
		}
		
		if(IsKeyDown(KEY_CTRL_DOWN)) {
			skt.forceY += FORCE_STEP;
			if(skt.forceY > FORCE_MAX) skt.forceY = FORCE_MAX;
		}
		else if(IsKeyDown(KEY_CTRL_UP)) {
			skt.forceY -= FORCE_STEP;
			if(skt.forceY < -FORCE_MAX) skt.forceY = -FORCE_MAX;
		}
		else if(skt.forceY > 0) {
			skt.forceY += FORCE_STEP2;
			if(skt.forceY < 0) skt.forceY = 0;
		}
		else if(skt.forceY < 0) {
			skt.forceY -= FORCE_STEP2;
			if(skt.forceY > 0) skt.forceY = 0;
		}
		
		moveSkater(&skt);
		
		if(RTC_getTicks()-timeSkating > TIME_SKATING) {
			timeSkating = RTC_getTicks();
			posSkating = !posSkating;
		}
		
		ML_clear_vram();
		
		//ice
		if(iLoop == 0 && i == 0) {
			traceX[i] = (int)skt.skaterX+5;
			traceY[i] = (int)skt.skaterY+14;
			i++;
		}
		else {
			i1 = i-1; i2 = i; 
			if(i1 < 0) i1 = NB_TRACE-1;
			
			if((skt.skaterX+5-traceX[i1])*(skt.skaterX+5-traceX[i1])+(skt.skaterY+14-traceY[i1])*(skt.skaterY+14-traceY[i1]) >= DIFF_MIN_C) {
				traceX[i] = (int)skt.skaterX+5;
				traceY[i] = (int)skt.skaterY+14;
				i++;
				if(i >= NB_TRACE) {
					i = 0;
					iLoop = 1;
				}
			}
			
			defend1 = i2;
			if(defend1 == 0) defend1 = NB_TRACE;
			defend2 = i2+1;
			for(j = 1 ; j <= NB_TRACE ; j++) {
				if((iLoop || j < i2) && j != defend1 && j != defend2) {
					j1 = j-1; j2 = j;
					if(j == NB_TRACE) j2 = 0;
					ML_line(traceX[j1], traceY[j1], traceX[j2], traceY[j2], ML_BLACK);
					
					if(!hole) {
						if(intersec(traceX[j1], traceY[j1], traceX[j2], traceY[j2], traceX[i1], traceY[i1], traceX[i2], traceY[i2])) {
							cpt = 1;
							k = j1;
							while(k != i2) {
								k++; cpt++;
								if(k >= NB_TRACE) k = 0;
							}
							
							if(cpt > CPT_MIN) {
								k = j1;
								cpt2 = 0;
								for(cpt ; cpt > 0 ; cpt -= 5) {
									if(traceX[k]>=0 && traceX[k]<=127 && traceY[k]>=0 && traceY[k]<=63) {
										abscissa[cpt2] = traceX[k];
										ordinate[cpt2] = traceY[k];
										cpt2++;
										k += 5;
										if(k >= NB_TRACE) k -= NB_TRACE;
									}
								}
								hole = cpt2-1;
								timeHole = RTC_getTicks();
								for(k = 0 ; k < NB_TEST ; k++) {
									if(RTC_getTicks()-timeEnemy > TIME_ENEMY && inHole(enemyX+2+random(0, 7), enemyY+6+random(0, 9), &abscissa, &ordinate, hole)) {
										nbEnemy++;
										do {
											enemyX = 15+random(0, 86);
											enemyY = 16+random(0, 16);
										} while((enemyX-skt.skaterX)*(enemyX-skt.skaterX)+(enemyY-skt.skaterY)*(enemyY-skt.skaterY) < ENEMY_SIZE_C);
										hole = 0;
										timeEnemy = RTC_getTicks();
										k = NB_TEST;
									}
								}
							}
						}
					}
				}
			}
			if(hole) {
				polygon(&abscissa, &ordinate, hole);
				if(RTC_getTicks()-timeHole > TIME_HOLE) hole = 0;
				else if(RTC_getTicks()-timeHole > TIME_HOLE_MIN && inHole((int)skt.skaterX+5, (int)skt.skaterY+14, &abscissa, &ordinate, hole)) {
					return -nbEnemy-1;
				}
			}
			ML_line(traceX[i1], traceY[i1], (int)skt.skaterX+5, (int)skt.skaterY+14, ML_BLACK);
		}
		//end ice
		
		if(skt.speedX != 0 || skt.speedY != 0) {
			if(skt.speedX*skt.speedX > skt.speedY*skt.speedY) {
				if(skt.speedX > 0) posSkater = 2;
				else posSkater = 3;
			}
			else {
				if(skt.speedY > 0) posSkater = 0;
				else posSkater = 1;
			}
		}
		
		if(enemyY <= skt.skaterY) {
			ML_bmp_and(enemy->tile[0], enemyX, enemyY, 11, 16);
			ML_bmp_or(enemy->tile[1], enemyX, enemyY, 11, 16);
		}
		if(posSkater <= 1) {
			ML_bmp_and(sktFrontBack->tile[4*posSkater+2*posSkating], (int)skt.skaterX, (int)skt.skaterY, 11, 15);
			ML_bmp_or(sktFrontBack->tile[4*posSkater+1+2*posSkating], (int)skt.skaterX, (int)skt.skaterY, 11, 15);
		}
		else {
			ML_bmp_and(sktSide->tile[4*(posSkater-2)+2*posSkating], (int)skt.skaterX, (int)skt.skaterY, 10, 15);
			ML_bmp_or(sktSide->tile[4*(posSkater-2)+1+2*posSkating], (int)skt.skaterX, (int)skt.skaterY, 10, 15);
		}
		if(enemyY > skt.skaterY) {
			ML_bmp_and(enemy->tile[0], enemyX, enemyY, 11, 16);
			ML_bmp_or(enemy->tile[1], enemyX, enemyY, 11, 16);
		}
		
		
		ML_display_vram();
	}
	return nbEnemy;
}

int intersec(int a0, int a1, int b0, int b1, int c0, int c1, int d0, int d1) {
	int r, s, den;
	
	den = (b0-a0)*(d1-c1)-(b1-a1)*(d0-c0);
	
	if(den == 0) return 0;
	
	r = (a1-c1)*(d0-c0)-(a0-c0)*(d1-c1)/den;
	s = (a1-c1)*(b0-a0)-(a0-c0)*(b1-a1)/den;
	
	if(r >= 0 && r <= 1 && s >= 0 && s <= 1) return 1;
	else return 0;
}

void polygon(int *x, int *y, int cpt) {
	int i;

	for(i = 1 ; i <= cpt ; i++) {
			ML_line(x[i-1], y[i-1], x[i], y[i], ML_BLACK);
	}
		ML_line(x[cpt-1], y[cpt-1], x[0], y[0], ML_BLACK);
}

int inHole(int x, int y, int *holeX, int *holeY, int n) {
	int i, j;
	int bool = 0;
	
	j = 0;
	for(i = 1 ; i <= n ; i++) {
		if(intersec(x, y, x, 0, holeX[i-1], holeY[i-1], holeX[i], holeY[i])) j++;
	}
	if(j) bool++;
	
	j = 0;
	for(i = 1 ; i <= n ; i++) {
		if(intersec(x, y, x, 63, holeX[i-1], holeY[i-1], holeX[i], holeY[i])) j++;
	}
	if(j) bool++;
	
	j = 0;
	for(i = 1 ; i <= n ; i++) {
		if(intersec(x, y, 0, y, holeX[i-1], holeY[i-1], holeX[i], holeY[i])) j++;
	}
	if(j) bool++;
	
	j = 0;
	for(i = 1 ; i <= n ; i++) {
		if(intersec(x, y, 127, y, holeX[i-1], holeY[i-1], holeX[i], holeY[i])) j++;
	}
	if(j) bool++;
	
	if(bool >= 4) return 1;
	else return 0;
}